from enum import Enum
import re


def validate_hostname_and_return_error_message(hostname):
    if '/' in hostname or '\\' in hostname:
        return 'The hostname can not contain slashes. Please only use the domain'

    if re.findall(r'[\s]', hostname):
        return 'The hostname can not contain whitespace'

    if hostname.split('.')[-1].isdigit():
        return 'The hostname can not be an IP address'

    # Using 'in' to catch subdomains as well i.e. 'account.biolib.com'
    if 'biolib.com' in hostname:
        return 'The biolib.com domain can not be used as a remote host'


stdout_render_types = [
    ('text', 'Text'),
    ('markdown', 'Markdown'),
]


class ChoicesEnum(Enum):
    """
    Class to be inherited from when using enums with the choice option in CharFields
    """
    @classmethod
    def choices(cls):
        return tuple((i.value, i.name) for i in cls)

    @classmethod
    def values(cls):
        return [i.value for i in cls]


class AllowedYAMLEnvironments(ChoicesEnum):
    LOCAL_DOCKER = 'local-docker'
    DOCKERHUB = 'dockerhub'
    BIOLIB_APP = 'biolib'


def normalize(string):
    return string.replace('-', '_').lower()


custom_executors = {
    'python3': {
        'latest': '1.0.0',
        'versions': [
            '1.0.0'
        ],
    },
    'onnx': {
        'latest': '1.0.0',
        'versions': [
            '1.0.0'
        ],
    },
    'r': {
        'latest': '1.0.0',
        'versions': [
            '1.0.0'
        ],
    },
    'rust': {
        'latest': '1.0.0',
        'versions': [
            '1.0.0'
        ],
    },
    'tensorflow': {
        'latest': '1.0.0',
        'versions': [
            '1.0.0'
        ],
    },
    'tflite': {
        'latest': '1.0.0',
        'versions': [
            '1.0.0'
        ],
    },
    'emscripten': {
        'latest': '1.0.0',
        'versions': [
            '1.0.0'
        ],
    }
}

old_to_new_executors_map = {
    'bl-python3': 'python3',
    'bl-onnx': 'onnx',
    'bl-rust': 'rust',
    'bl-r': 'r',
    'bl-tensorflow': 'tensorflow',
    'bl-emscripten': 'emscripten',
    'bl-tflite': 'tflite'
}




from urllib.parse import urlparse

from biolib import utils
from biolib.biolib_api_client import BiolibApiClient
from biolib.biolib_api_client.biolib_job_api import BiolibJobApi
from biolib.biolib_logging import logger, logger_no_user_data
from biolib.compute_node.job_worker.executors.types import RemoteExecuteOptions

from biolib.compute_node.job_worker.job_storage import JobStorage


class RemoteExecutor:

    @staticmethod
    def execute_job(options: RemoteExecuteOptions, module_input_serialized: bytes) -> bytes:
        job_id = options['job']['public_id']
        cloud_job = BiolibJobApi.create_cloud_job(module_name='main', job_id=job_id)
        logger.debug(f"Cloud: Job created with id {cloud_job['public_id']}")

        node_url = cloud_job['compute_node_info']['url']
        if utils.BIOLIB_CLOUD_BASE_URL:
            logger_no_user_data.debug('Using cloud proxy URL from env var BIOLIB_CLOUD_BASE_URL')
            node_url = utils.BIOLIB_CLOUD_BASE_URL + urlparse(node_url).path

        logger_no_user_data.debug(f'Using compute node URL "{node_url}"')

        if utils.BASE_URL_IS_PUBLIC_BIOLIB:
            aes_key_string_b64 = JobStorage.generate_and_store_key_buffer_for_job(job_id)
            BiolibJobApi.start_cloud_job(job_id, module_input_serialized, node_url, aes_key_string_b64)
        else:
            BiolibJobApi.start_cloud_job(job_id, module_input_serialized, node_url)

        utils.STREAM_STDOUT = True
        BiolibJobApi.await_compute_node_status(
            compute_type='Cloud',
            node_url=node_url,
            retry_interval_seconds=1.5,
            retry_limit_minutes=10080,  # 1 Week
            status_to_await='Result Ready',
            job=options['job'],
        )

        BiolibApiClient.refresh_auth_token()
        return JobStorage.get_result(job=options['job'])



import logging
import sys
import os

_DEFAULT_LOGGER_FORMAT = '%(asctime)s | %(levelname)s : %(message)s'

# define global logging format
logging.basicConfig(format=_DEFAULT_LOGGER_FORMAT, level=logging.INFO, stream=sys.stdout)

# define extra log levels
TRACE = 5
logging.addLevelName(TRACE, 'TRACE')


# note: Logger classes should never be instantiated directly
class _BioLibLogger(logging.Logger):

    def __init__(self, name: str, level=logging.INFO):
        super(_BioLibLogger, self).__init__(name=name, level=level)

    def configure(self, default_log_level):
        env_log_level = os.getenv('BIOLIB_LOG')
        if env_log_level is None or env_log_level == '':
            self.setLevel(default_log_level)
        else:
            self.setLevel(env_log_level.upper())

    def setLevel(self, level) -> None:
        try:
            normalized_level = level.upper() if isinstance(level, str) else level
            super(_BioLibLogger, self).setLevel(normalized_level)
        except ValueError:
            raise Exception(f'Unknown log level "{level}"') from None

        global_root_logger = logging.getLogger()
        # only activate debug logging globally if user selected trace logging
        if self.level == TRACE:
            global_root_logger.setLevel(logging.DEBUG)
        elif self.level == logging.DEBUG:
            global_root_logger.setLevel(logging.INFO)
        else:
            global_root_logger.setLevel(self.level)


def _get_biolib_logger_instance(name: str) -> _BioLibLogger:
    # for thread safety use the global lock of logging
    logging._acquireLock()  # type: ignore # pylint: disable=protected-access

    original_logger_class = logging.getLoggerClass()
    try:
        # change logger class temporarily to get instance of _BioLibLogger
        logging.setLoggerClass(_BioLibLogger)
        biolib_logger = logging.getLogger(name=name)
        # change the logger class back to original so we do not interfere with other libraries
        logging.setLoggerClass(original_logger_class)
        return biolib_logger  # type: ignore
    finally:
        logging._releaseLock()  # type: ignore # pylint: disable=protected-access


def _get_no_user_data_logger() -> _BioLibLogger:
    _logger_no_user_data = _get_biolib_logger_instance(name='biolib_no_user_data')

    # TODO: Simplify by refactoring to env BIOLIB_ENVIRONMENT_IS_CLOUD: boolean
    if os.getenv('BIOLIB_CLOUD_ENVIRONMENT', '').lower() == 'non-enclave':
        handler = logging.FileHandler(filename='/tmp/biolib_no_user_data.log')
        formatter = logging.Formatter(_DEFAULT_LOGGER_FORMAT)
        handler.setFormatter(formatter)
        _logger_no_user_data.addHandler(handler)

    return _logger_no_user_data


# expose loggers
logger = _get_biolib_logger_instance(name='biolib')
logger_no_user_data = _get_no_user_data_logger()
